import logging

logger = logging.getLogger(__name__)

import os

xla_env = os.environ.get("XLA_FLAGS")

xla_env_set = True

if xla_env is None:
    xla_env_set = False
elif isinstance(xla_env, str):
    xla_env_set = "--xla_disable_hlo_passes=constant_folding" in xla_env

if not xla_env_set:
    logger.info(
        """
        For fast JAX compile times, the envirment variable XLA_FLAGS must be set to "--xla_disable_hlo_passes=constant_folding",
        which is currently not.
        
        In Python, to do this manually, use the code: 
        
        import os
        os.environ["XLA_FLAGS"] = "--xla_disable_hlo_passes=constant_folding"
        
        The environment variable has been set automatically for you now, however if JAX has already been imported, 
        this change will not take effect and JAX function compiling times may be slow. 
        
        Therefore, it is recommended to set this environment variable before running your script, e.g. in your terminal.
        """)

    os.environ['XLA_FLAGS'] = "--xla_disable_hlo_passes=constant_folding"

jax_enable_x64 = os.environ.get("JAX_ENABLE_X64")

if jax_enable_x64 is None:
    jax_enable_x64 = False
elif isinstance(jax_enable_x64, str):
    jax_enable_x64 = jax_enable_x64.lower() == "true"

if not jax_enable_x64:

    os.environ["JAX_ENABLE_X64"] = "True"

    logger.info(
        """"
        JAX 64-bit precision has been automatically enabled for you (JAX_ENABLE_X64=True),
        as double precision is required for most scientific computing applications.
        
        To enable 64 precision as default in JAX, set the environment variable 
        JAX_ENABLE_X64=true before running your script.
        """
    )



#!/usr/bin/env python
"""
Converts the configuration files and directories in a given directory into YAML configs.

Usage:
./convert_config.py path/to/directory
"""

import os
import shutil
import sys
from pathlib import Path

import yaml

from autoconf.directory_config import RecursiveConfig, YAMLConfig

target_path = Path(sys.argv[1])

config = RecursiveConfig(str(target_path))

for key in config.keys():
    value = config[key]
    if isinstance(value, YAMLConfig):
        continue

    d = value.dict()
    path = target_path / key
    with open(path.with_suffix(".yaml"), "w") as f:
        yaml.dump(d, f)

    try:
        os.remove(path.with_suffix(".ini"))
    except FileNotFoundError:
        pass

    shutil.rmtree(path, ignore_errors=True)



#!/usr/bin/env python
"""
Converts JSON prior configs to YAML equivalent.

Usage:
./convert_prior_configs.py /path/to/prior/directory
"""
import json
import os
import sys
from pathlib import Path

import oyaml as yaml

ORDER = [
    "type",
    "mean",
    "sigma",
    "lower_limit",
    "upper_limit",
    "width_modifier",
    "limits",
]

for path in Path(sys.argv[1]).rglob("*.json"):
    with open(path) as f:
        d = json.load(f)

    with open(path.with_suffix(".yaml"), "w") as f:
        yaml.dump(d, f)

    os.remove(path)


def sort_dict(obj):
    if isinstance(obj, dict):
        return {
            key: sort_dict(value)
            for key, value in sorted(
                obj.items(),
                key=lambda item: ORDER.index(item[0]) if item[0] in ORDER else 999,
            )
        }
    if isinstance(obj, list):
        return list(map(sort_dict, obj))
    return obj


for path in Path(sys.argv[1]).rglob("*.yaml"):
    with open(path) as f:
        d = yaml.safe_load(f)

    print(sort_dict(d))
    with open(path, "w") as f:
        yaml.dump(sort_dict(d), f)



import inspect
import json
import logging
import os

import numpy as np
from pathlib import Path
from typing import Union, Callable, Set, Tuple

from autoconf.class_path import get_class_path, get_class

logger = logging.getLogger(__name__)

np_type_map = {
    "bool": "bool_",
}


def nd_array_as_dict(obj: np.ndarray) -> dict:
    """
    Converts a numpy array to a dictionary representation.
    """
    np_type = str(obj.dtype)
    return {
        "type": "ndarray",
        "array": obj.tolist(),
        "dtype": np_type_map.get(np_type, np_type),
    }


def nd_array_from_dict(nd_array_dict: dict, **_) -> np.ndarray:
    """
    Converts a dictionary representation back to a numpy array.
    """
    return np.array(nd_array_dict["array"], dtype=getattr(np, nd_array_dict["dtype"]))


def is_array(obj) -> bool:
    """
    True if the object is a numpy array or an ArrayImpl (i.e. from JAX)
    """
    if isinstance(obj, np.ndarray):
        return True
    try:
        return obj.__class__.__name__ == "ArrayImpl"
    except AttributeError:
        return False


def compound_key_dict(obj):
    """
    Converts a dictionary with compound keys to a dictionary with a single key.
    """
    return {
        "type": "compound_dict",
        "arguments": [
            {
                "key": to_dict(key),
                "value": to_dict(value),
            }
            for key, value in obj.items()
        ],
    }


def to_dict(obj, filter_args: Tuple[str, ...] = ()) -> dict:
    if hasattr(obj, "dict"):
        try:
            return obj.dict()
        except TypeError as e:
            logger.debug(e)

    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj

    if isinstance(obj, slice):
        return {
            "type": "slice",
            "start": to_dict(obj.start),
            "stop": to_dict(obj.stop),
            "step": to_dict(obj.step),
        }

    if isinstance(obj, np.number):
        return {
            "type": "np.number",
            "dtype": str(obj.dtype),
            "value": obj.item(),
        }

    if inspect.isfunction(obj):
        return {
            "type": "function",
            "class_path": obj.__module__ + "." + obj.__qualname__,
        }

    if is_array(obj):
        try:
            return nd_array_as_dict(obj)
        except Exception as e:
            logger.info(e)

    if isinstance(obj, Path):
        return {
            "type": "path",
            "path": str(obj),
        }

    if inspect.isclass(obj):
        return {
            "type": "type",
            "class_path": get_class_path(obj),
        }

    if isinstance(obj, list):
        return {"type": "list", "values": list(map(to_dict, obj))}
    if isinstance(obj, tuple):
        return {"type": "tuple", "values": list(map(to_dict, obj))}
    if isinstance(obj, dict):
        if any(
            not (isinstance(key, (str, int, float, bool)) or key is None)
            for key in obj.keys()
        ):
            return compound_key_dict(obj)

        return {
            "type": "dict",
            "arguments": {
                key: to_dict(value)
                for key, value in obj.items()
                if key not in filter_args
            },
        }

    if obj.__class__.__name__ == "method":
        return to_dict(obj())
    if obj.__class__.__module__ == "builtins":
        return obj

    if inspect.isclass(type(obj)):
        return instance_as_dict(obj, filter_args=filter_args)

    return obj


def get_arguments(obj) -> Set[str]:
    """
    Get the arguments of a class. This is done by inspecting the constructor.

    If the constructor has a **kwargs parameter, the arguments of the base classes are also included.

    Parameters
    ----------
    obj
        The class to get the arguments of.

    Returns
    -------
    A set of the arguments of the class.
    """
    args_spec = inspect.getfullargspec(obj.__init__)
    args = set(args_spec.args[1:])
    if args_spec.varkw:
        for base in obj.__bases__:
            if base is object:
                continue
            args |= get_arguments(base)
    return args


def instance_as_dict(obj, filter_args: Tuple[str, ...] = ()):
    """
    Convert an instance of a class to a dictionary representation.

    Serialises any children of the object which are given as constructor arguments
    or included in the __identifier_fields__ attribute.

    Sets any fields in the __nullify_fields__ attribute to None.

    Parameters
    ----------
    obj
        The instance of the class to be converted to a dictionary representation.
    filter_args
        A tuple of arguments to exclude from the dictionary representation.

    Returns
    -------
    A dictionary representation of the instance.
    """
    arguments = get_arguments(type(obj))
    try:
        arguments |= set(obj.__identifier_fields__)
    except (AttributeError, TypeError):
        pass

    argument_dict = {
        arg: getattr(obj, arg)
        for arg in arguments
        if arg not in filter_args
        if hasattr(obj, arg)
        and not inspect.ismethod(
            getattr(obj, arg),
        )
    }
    try:
        for field in obj.__nullify_fields__:
            argument_dict[field] = None
    except (AttributeError, TypeError):
        pass

    try:
        for field in obj.__exclude_fields__:
            try:
                argument_dict.pop(field)
            except KeyError:
                logger.debug(f"Field {field} not found in object")
    except (AttributeError, TypeError):
        pass

    return {
        "type": "instance",
        "class_path": get_class_path(obj.__class__),
        "arguments": {key: to_dict(value) for key, value in argument_dict.items()},
    }


__parsers = {
    "ndarray": nd_array_from_dict,
}


def register_parser(type_: str, parser: Callable[[dict], object]):
    """
    Register a parser for a given type.

    This parser will be used to instantiate objects of the given type from a
    dictionary representation.

    Parameters
    ----------
    type_
        The type of the object to be parsed. This is a string uniquely
        identifying the type.
    parser
        A function which takes a dictionary representation of an object and
        returns an instance of the object.
    """
    __parsers[type_] = parser


def from_dict(dictionary, **kwargs):
    """
    Instantiate an instance of a class from its dictionary representation.

    Parameters
    ----------
    dictionary
        An object which may be a dictionary representation of an object.

        This may contain the following keys:
        type: str
            The type of the object. This may be a built-in type, a numpy array,
            a list, a dictionary, a class, or an instance of a class.

            If a parser has been registered for the given type that parser will
            be used to instantiate the object.
        class_path: str
            The path to the class of the object. This is used to instantiate
            the object if it is not a built-in type.
        arguments: dict
            A dictionary of arguments to pass to the class constructor.

    Returns
    -------
    An object that was represented by the dictionary.
    """
    if isinstance(dictionary, (int, float, str, bool, type(None))):
        return dictionary

    if isinstance(dictionary, list):
        return list(map(from_dict, dictionary))

    if isinstance(dictionary, tuple):
        return tuple(map(from_dict, dictionary))

    try:
        type_ = dictionary["type"]
    except KeyError:
        logger.debug("No type field in dictionary")
        return dictionary
    except TypeError as e:
        logger.debug(e)
        return None

    if type_ == "path":
        return Path(dictionary["path"])

    if type_ == "slice":
        return slice(
            from_dict(dictionary["start"]),
            from_dict(dictionary["stop"]),
            from_dict(dictionary["step"]),
        )

    if type_ == "np.number":
        return getattr(
            np,
            dictionary["dtype"],
        )(dictionary["value"])

    if type_ == "function":
        return get_class(dictionary["class_path"])

    if type_ in __parsers:
        return __parsers[type_](dictionary, **kwargs)

    if type_ == "list":
        return list(map(from_dict, dictionary["values"]))
    if type_ == "tuple":
        return tuple(map(from_dict, dictionary["values"]))
    if type_ == "dict":
        return {
            key: from_dict(value, **kwargs)
            for key, value in dictionary["arguments"].items()
        }
    if type_ == "compound_dict":
        return {
            from_dict(item["key"], **kwargs): from_dict(item["value"], **kwargs)
            for item in dictionary["arguments"]
        }

    if type_ == "type":
        return get_class(dictionary["class_path"])

    cls = get_class(dictionary["class_path"])

    if cls is np.ndarray:
        return nd_array_from_dict(dictionary)
    if hasattr(cls, "from_dict"):
        return cls.from_dict(dictionary, **kwargs)

    # noinspection PyArgumentList
    return cls(
        **{
            name: from_dict(value, **kwargs)
            for name, value in dictionary["arguments"].items()
        }
    )


def from_json(file_path: str):
    """
    Load the dictable object to a .json file, whereby all attributes are converted from the .json file's dictionary
    representation to create the instance of the object

    A json file of the instance can be created from the .json file via the `output_to_json` method.

    Parameters
    ----------
    file_path
        The path to the .json file that the dictionary representation of the object is loaded from.
    """
    with open(file_path, "r+") as f:
        cls_dict = json.load(f)

    return from_dict(cls_dict)


def output_to_json(obj, file_path: Union[Path, str]):
    """
    Output the dictable object to a .json file, whereby all attributes are converted to a dictionary representation
    first.

    An instance of the object can be created from the .json file via the `from_json` method.

    Parameters
    ----------
    file_path
        The path to the .json file that the dictionary representation of the object is written too.
    """
    file_path = Path(file_path)
    file_dir = Path(*file_path.parts[:-1])
    file_dir.mkdir(parents=True, exist_ok=True)

    with open(file_path, "w+") as f:
        json.dump(to_dict(obj), f, indent=4)



#!/usr/bin/env python
"""
Generate prior configuration for all classes in a package, recursively
"""

from sys import argv

from autoconf.json_prior import generate

if __name__ == "__main__":
    try:
        generate.generate(argv[1])
    except KeyError:
        print("Usage: ./generate_priors.py /path/to/project")



import logging
import os
os.environ['XLA_FLAGS'] = "--xla_disable_hlo_passes=constant_folding"

logger = logging.getLogger(__name__)

def check_jax_using_gpu(log_gpu_warning : bool = True) -> bool:

    import jax

    devices = jax.devices()

    # Normalize device_kind for matching
    for d in devices:
        kind = str(d).lower()

        no_gpu = False

        if "gpu" in kind or "cuda" in kind:
            print(f"GPU / CUDA detected: {d}")
        elif "tpu" in kind:
            print(f"TPU detected: {d}")
        elif "cpu" in kind:
            print(f"CPU detected: {d}")
            no_gpu = True
        else:
            print(f"Other device detected: {d}")
            no_gpu = True

    if no_gpu and log_gpu_warning:

        logger.info(
            """
            JAX did not detect a GPU or TPU device and is using the CPU for computations.

            PyAutoLens runs > 50 times faster with a GPU, so it is recommended to reinstall
            JAX With GPU support, if you have a GPU available, by following the JAX 
            installation instructions.
            """)

    return no_gpu

def for_autolens(raise_error_if_not_gpu: bool = True):

    import os
    os.environ['XLA_FLAGS'] = "--xla_disable_hlo_passes=constant_folding"

    import subprocess
    import sys

    try:

        import google.colab

        no_gpu = check_jax_using_gpu(log_gpu_warning=False)

        if no_gpu and raise_error_if_not_gpu:

            raise RuntimeError(
                """
                No GPU detected in Google Colab. PyAutoLens runs > 50 times faster with a GPU, so switch GPU 
                on in Colab settings.

                To do this:

                - Click "Runtime" in the top menu.
                - Click "Change runtime type".
                - Under "Hardware accelerator", select one of the "GPU" options available.

                You can set up Colab with a CPU (e.g. if GPUs are unavailable)  by uncommenting the 
                line "raise_error_if_not_gpu=False" in the setup_colab() function call.
                """
            )

    except ImportError:
        print(
            """
            You are not running in a Google Colab environment so cannot use the setup_colab() function.

            You should therefore have PyAutoLens installed locally in your environment already (e.g. via pip or 
            conda and can run the rest of your script normally).
            
            You may now continue running your script or Notebook.
            """
        )
        return

    print()
    print("Now Installing PyAutoLens and setting up Colab Environment:")

    # Install required packages
    subprocess.check_call([sys.executable, "-m", "pip", "install",
                           "autoconf", "autofit", "autoarray", "autogalaxy", "autolens",
                           "pyvis==0.3.2", "dill==0.4.0", "jaxnnls",
                           "pyprojroot==0.2.0", "nautilus-sampler==1.0.4",
                           "timeout_decorator==0.5.0", "anesthetic==2.8.14",
                           "--no-deps"])

    from autoconf import conf

    try:
        subprocess.run([
            "git", "clone", "https://github.com/Jammy2211/autolens_workspace"
        ], check=True)
    except subprocess.CalledProcessError as e:
        print("Workspace already exists so not cloning again.")

    os.chdir("/content/autolens_workspace")

    conf.instance.push(
        new_path="/content/autolens_workspace/config",
        output_path="/content/autolens_workspace/output",
    )

    print(
        """
        ***Google Colab Setup Complete, which included:***

        - Installation of PyAutoLens and other required packages.
        - Cloning of the autolens_workspace GitHub repository.
        - Setting up environment variables for JAX for improved performance.
        - Setting the configuration paths to the workspace config and output folders suitable for Colab.
        """
    )


#!/usr/bin/env python

from configparser import ConfigParser
from sys import argv

from autofit.tools import edenise


def main(root_directory):
    try:
        config = ConfigParser()
        config.read(f"{root_directory}/eden.ini")

        edenise.edenise(
            root_directory, config.get("eden", "name"), config.get("eden", "prefix")
        )
    except ValueError:
        print("Usage: ./edenise.py root_directory")
        exit(1)


if __name__ == "__main__":
    main(argv[1])



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    try:
        from astropy.io import fits
    except ImportError:
        pass

try:
    from astropy.io import fits
except ImportError:
    pass

import numpy as np
from pathlib import Path
from typing import Dict, Optional, Union, List

from autoconf import conf


def flip_for_ds9_from(values: np.ndarray) -> np.ndarray:
    """
    Returns the input 2D array flipped upside-down depending on the project config files.

    This is for Astronomy projects so that structures appear the same orientation as `.fits` files loaded in DS9.

    Parameters
    ----------
    values
        The 2D array that is flipped upside-down.

    Returns
    -------
    The 2D array flipped upside-down.

    Examples
    --------
    data = np.ones((5,5))

    flip_for_ds9_from(data)
    """
    if len(values.shape) > 1:

        flip_for_ds9 = conf.instance["general"]["fits"]["flip_for_ds9"]

        if flip_for_ds9:
            return np.flipud(values)

    return values

def hdu_list_for_output_from(
    values_list: List[np.ndarray],
    header_dict: Optional[dict] = None,
    ext_name_list: Optional[List[str]] = None,
) -> fits.HDUList:
    """
    Returns the HDU list which can be used to output arrays to a .fits file.

    Before outputting each NumPy array, the array may be flipped upside-down using np.flipud depending on the project
    config files. This is for Astronomy projects so that structures appear the same orientation as ``.fits`` files
    loaded in DS9.

    The output .fits files may contain multiple HDUs comprising different images. Conventionally, the first array,
    the `PrimaryHDU`, contains the 2D mask applied to the data and the remaining HDUs contain the data itself.
    The mask is used to add information to the header, for example the pixel scale of the data.
    
    Each HDU contains its `ext_name` in the header, which is visible when the .fits file is loaded in DS9.

    Parameters
    ----------
    values
        The 1D or 2D array that is written to fits.
    header_dict
        A dictionary of values that are written to the header of the .fits file.
    ext_name_list
        The names of the extension in the fits file, which displays in the header of the fits file and is visible
        for example when the .fits is loaded in DS9.

    Returns
    -------
    The HDU list containing the data and its header which can then be written to .fits.

    Examples
    --------
    data = np.ones((5,5))
    noise_map = np.ones((5,5))

    hdu_list_for_output_from(
        values_list=[data, noise_map]
        header_dict={"Example": 0.5},
        ext_name_list=["data", "noise_map"]
    )
    """
    hdu_list = []
    
    header = fits.Header()

    if header_dict is not None:
        for key, value in header_dict.items():
            # Convert enum to its string value if needed
            key_str = key.value if isinstance(key, Enum) else key
            try:
                header.append((key_str, value, [""]))
            except ValueError:
                header.append((key_str, float(value), [""]))

    for i, values in enumerate(values_list):
    
        if ext_name_list is not None:
            header["EXTNAME"] = ext_name_list[i].upper()

        # Convert from JAX
        try:
            values = np.array(values.array)
        except AttributeError:
            values = np.array(values)

        values = flip_for_ds9_from(values)
                
        if i == 0:
            hdu_list.append(fits.PrimaryHDU(values, header=header))
        else:
            hdu_list.append(fits.ImageHDU(values, header=header))
         
    return fits.HDUList(hdus=hdu_list)

def output_to_fits(
    values: np.ndarray,
    file_path: Union[Path, str],
    overwrite: bool = False,
    header_dict: Optional[dict] = None,
    ext_name: Optional[str] = None,
):
    """
    Write a NumPy array to a .fits file.

    Before outputting a NumPy array, the array may be flipped upside-down using np.flipud depending on the project
    config files. This is for Astronomy projects so that structures appear the same orientation as ``.fits`` files
    loaded in DS9.

    Parameters
    ----------
    values
        The numpy array of values that is written to fits.
    file_path
        The full path of the file that is output, including the file name and ``.fits`` extension.
    overwrite
        If `True` and a file already exists with the input file_path the .fits file is overwritten. If `False`, an
        error is raised.
    header_dict
        A dictionary of values that are written to the header of the .fits file.
    ext_name
        The name of the extension in the fits file, which displays in the header of the fits file and is visible.

    Examples
    --------
    values = np.ones((5,5))
    numpy_array_to_fits(values=values, file_path='/path/to/file/filename.fits', overwrite=True)
    """

    file_path = Path(file_path)

    file_dir = Path(*file_path.parts[:-1])
    file_dir.mkdir(parents=True, exist_ok=True)

    if overwrite and file_path.is_file():
        file_path.unlink()

    hdu = hdu_list_for_output_from(
        values_list=[values],
        header_dict=header_dict,
        ext_name_list=[ext_name] if ext_name is not None else None,
    )

    hdu.writeto(file_path)

def ndarray_via_hdu_from(hdu):
    """
    Returns an ``Array2D`` by from a `PrimaryHDU` object which has been loaded via `astropy.fits`

    This assumes that the `header` of the `PrimaryHDU` contains an entry named `PIXSCALE` which gives the
    pixel-scale of the array.

    For a full description of ``Array2D`` objects, including a description of the ``slim`` and ``native`` attribute
    used by the API, see
    the :meth:`Array2D class API documentation <autoarray.structures.arrays.uniform_2d.AbstractArray2D.__new__>`.

    Parameters
    ----------
    primary_hdu
        The `PrimaryHDU` object which has already been loaded from a .fits file via `astropy.fits` and contains
        the array data and the pixel-scale in the header with an entry named `PIXSCALE`.
    origin
        The (y,x) scaled units origin of the coordinate system.

    Examples
    --------

    .. code-block:: python

        from astropy.io import fits
        import autoarray as aa

        primary_hdu = fits.open("path/to/file.fits")

        array_2d = aa.Array2D.from_primary_hdu(
            primary_hdu=primary_hdu,
        )
    """
    values = hdu.data.astype("float")
    return flip_for_ds9_from(values)


def ndarray_via_fits_from(
    file_path: Union[Path, str], hdu: int, do_not_scale_image_data: bool = False
):
    """
    Read a 2D NumPy array from a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path
        The full path of the file that is loaded, including the file name and ``.fits`` extension.
    hdu
        The HDU extension of the array that is loaded from the .fits file.
    do_not_scale_image_data
        If True, the .fits file is not rescaled automatically based on the .fits header info.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .fits file.

    Examples
    --------
    array_2d = ndarray_via_fits_from(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path, do_not_scale_image_data=do_not_scale_image_data)
    return ndarray_via_hdu_from(hdu_list[hdu])


def header_obj_from(file_path: Union[Path, str], hdu: int) -> Dict:
    """
    Read a 2D NumPy array from a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path
        The full path of the file that is loaded, including the file name and ``.fits`` extension.
    hdu
        The HDU extension of the array that is loaded from the .fits file.
    do_not_scale_image_data
        If True, the .fits file is not rescaled automatically based on the .fits header info.

    Returns
    -------
    dict
        The header dictionary.

    Examples
    --------
    array_2d = ndarray_via_fits_from(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path)
    return hdu_list[hdu].header








import builtins
import importlib
import re
from typing import List, Type


def get_class_path(cls: type) -> str:
    """
    The full import path of the type
    """
    if hasattr(cls, "__class_path__"):
        cls = cls.__class_path__
    return re.search("'(.*)'", str(cls))[1]


def get_class(class_path: str) -> Type[object]:
    return GetClass(class_path).cls


class GetClass:
    def __init__(self, class_path):
        self.class_path = class_path

    @property
    def _class_path_array(self) -> List[str]:
        """
        A list of strings describing the module and class of the
        real object represented here
        """
        return self.class_path.split(".")

    @property
    def _class_name(self) -> str:
        """
        The name of the real class
        """
        return self._class_path_array[-1]

    @property
    def _module_path(self) -> str:
        """
        The path of the module containing the real class
        """
        return ".".join(self._class_path_array[:-1])

    @property
    def _module(self):
        """
        The module containing the real class
        """
        try:
            return importlib.import_module(self._module_path)
        except ValueError:
            return builtins

    @property
    def cls(self) -> Type[object]:
        """
        The class of the real object
        """
        return getattr(self._module, self._class_name)



import configparser
import os
from abc import abstractmethod, ABC
from pathlib import Path

import yaml

from autoconf import exc


class AbstractConfig(ABC):
    @abstractmethod
    def _getitem(self, item):
        pass

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.items()[item]
        return self._getitem(item)

    def items(self):
        return [(key, self[key]) for key in self.keys()]

    def __len__(self):
        return len(self.items())

    @abstractmethod
    def keys(self):
        pass

    def family(self, cls):
        for cls in family(cls):
            key = cls.__name__
            try:
                return self[key]
            except (KeyError, configparser.NoOptionError):
                pass
        raise KeyError(f"No configuration found for {cls.__name__}")

    def dict(self):
        d = {}
        for key in self.keys():
            value = self[key]
            if isinstance(value, AbstractConfig):
                value = value.dict()
            d[key] = value
        return d


class DictConfig(AbstractConfig):
    def keys(self):
        return self.d.keys()

    def __init__(self, d):
        self.d = d

    def __getitem__(self, item):
        value = self.d[item]
        if isinstance(value, dict):
            return DictConfig(value)
        return value

    def _getitem(self, item):
        return self[item]

    def items(self):
        for key in self.d:
            yield key, self[key]


class YAMLConfig(AbstractConfig):
    def __init__(self, path):
        with open(path) as f:
            self._dict = yaml.safe_load(f)

    def _getitem(self, item):
        value = self._dict[item]
        if isinstance(value, dict):
            return DictConfig(value)
        return value

    def keys(self):
        return self._dict.keys()


class SectionConfig(AbstractConfig):
    def __init__(self, path, parser, section):
        self.path = path
        self.section = section
        self.parser = parser

    def keys(self):
        with open(self.path) as f:
            string = f.read()

        lines = string.split("\n")
        is_section = False
        for line in lines:
            if line == f"[{self.section}]":
                is_section = True
                continue
            if line.startswith("["):
                is_section = False
                continue
            if is_section and "=" in line:
                yield line.split("=")[0]

    def _getitem(self, item):
        try:
            result = self.parser.get(self.section, item)
            if result.lower() == "true":
                return True
            if result.lower() == "false":
                return False
            if result.lower() in ("none", "null"):
                return None
            if result.isdigit():
                return int(result)
            try:
                return float(result)
            except ValueError:
                return result
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise KeyError(f"No configuration found for {item} at path {self.path}")


class NamedConfig(AbstractConfig):
    def __init__(self, config_path):
        """
        Parses generic config

        Parameters
        ----------
        config_path
            The path to the config file
        """
        self.path = config_path
        self.parser = configparser.ConfigParser()
        self.parser.read(self.path)

    def keys(self):
        return self.parser.sections()

    def _getitem(self, item):
        return SectionConfig(
            self.path,
            self.parser,
            item,
        )


class RecursiveConfig(AbstractConfig):
    def keys(self):
        try:
            return [
                path.split(".")[0]
                for path in os.listdir(self.path)
                if all(
                    [
                        path != "priors",
                        len(path.split(".")[0]) != 0,
                        os.path.isdir(f"{self.path}/{path}")
                        or path.endswith(".ini")
                        or path.endswith(".yaml")
                        or path.endswith(".yml"),
                    ]
                )
            ]
        except FileNotFoundError as e:
            raise KeyError(f"No configuration found at {self.path}") from e

    def __init__(self, path):
        self.path = Path(path)

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path}>"

    def _getitem(self, item):
        item_path = self.path / f"{item}"
        file_path = f"{item_path}.ini"
        if os.path.isfile(file_path):
            return NamedConfig(file_path)
        yml_path = item_path.with_suffix(".yml")
        if yml_path.exists():
            return YAMLConfig(yml_path)
        yaml_path = item_path.with_suffix(".yaml")
        if yaml_path.exists():
            return YAMLConfig(yaml_path)
        if os.path.isdir(item_path):
            return RecursiveConfig(item_path)
        raise KeyError(f"No configuration found for {item} at path {self.path}")


class PriorConfigWrapper:
    def __init__(self, prior_configs):
        self.prior_configs = prior_configs

    def for_class_and_suffix_path(self, cls, path):
        for config in self.prior_configs:
            try:
                return config.for_class_and_suffix_path(cls, path)
            except KeyError:
                pass
        directories = " ".join(str(config.directory) for config in self.prior_configs)

        print()

        raise exc.ConfigException(
            f"No prior config found for class: \n\n"
            f"{cls.__name__} \n\n"
            f"For parameter name and path: \n\n "
            f"{'.'.join(path)} \n\n "
            f"In any of the following directories:\n\n"
            f"{directories}\n\n"
            f"Either add configuration for the parameter or a type annotation for a class with valid configuration.\n\n"
            f"The following readthedocs page explains prior configuration files in PyAutoFit and will help you fix "
            f"the error https://pyautofit.readthedocs.io/en/latest/general/adding_a_model_component.html"
        )


def family(current_class):
    yield current_class
    for next_class in current_class.__bases__:
        for val in family(next_class):
            yield val



import inspect
import json
import logging
from collections.abc import Sized
from pathlib import Path
from typing import List, Type, Tuple

import yaml

from autoconf.directory_config import family

logger = logging.getLogger(__name__)

default_prior = {
    "type": "Uniform",
    "lower_limit": 0.0,
    "upper_limit": 1.0,
    "width_modifier": {"type": "Absolute", "value": 0.2},
    "limits": {"lower": 0.0, "upper": 1.0},
}


def make_config_for_class(cls):
    path = path_for_class(cls)
    arg_spec = inspect.getfullargspec(cls)
    arguments = arg_spec.args[1:]
    defaults = list(reversed(arg_spec.defaults or list()))

    config = dict()
    for i, argument in enumerate(reversed(arguments)):
        if i < len(defaults):
            default = defaults[i]
            if isinstance(default, Sized):
                for j in range(len(default)):
                    config[f"{argument}_{j}"] = default_prior
                continue
        config[argument] = default_prior

    return path, config


def path_for_class(cls) -> List[str]:
    """
    A list describing the import path for a given class.

    Parameters
    ----------
    cls
        A class with some module path

    Returns
    -------
    A list of modules terminating in the name of a class
    """
    return f"{cls.__module__}.{cls.__name__}".split(".")


class JSONPriorConfig:
    def __init__(self, config_dict: dict, directory=None):
        """
        Parses configuration describing priors associated with classes.

        The path pointing to a class is the same as the path to import it.

        Paths can be strings with '.' as a delimiter.
        {"module.class": config}

        Else they can be a series of dictionary keys.
        {"module": {"class": config}}

        Or any combination thereof.

        Parameters
        ----------
        config_dict
            A dictionary describing the prior configuration for constructor arguments
            of different classes.
        """
        self.obj = config_dict
        self.directory = directory
        self._path_value_map = None

    @property
    def paths(self):
        return list(self.path_value_map.keys())

    @property
    def path_value_map(self) -> dict:
        """
        A dictionary matching every possible path to the configuration it points to.
        """
        if self._path_value_map is None:

            def get_path_values(obj):
                path_values = dict()
                if isinstance(obj, dict):
                    for key, value in obj.items():
                        path_values[key] = value
                        for path, path_value in get_path_values(value).items():
                            path_values[f"{key}.{path}"] = path_value

                return path_values

            self._path_value_map = get_path_values(self.obj)
        return self._path_value_map

    @property
    def path_value_tuples(self) -> List[Tuple[str, object]]:
        """
        Tuple pairs matching every possible path to the configuration it points to.
        These are ordered by key length with the longest key first.
        """
        return sorted(
            list(self.path_value_map.items()),
            key=lambda item: len(item[0]),
            reverse=True,
        )

    @classmethod
    def from_directory(cls, directory: str) -> "JSONPriorConfig":
        """
        Load JSONPriorConfiguration from a file.

        Parameters
        ----------
        directory
            The path to a file.

        Returns
        -------
        A configuration instance.
        """
        config_dict = dict()

        config_path = Path(directory)

        for suffix, parser in [
            ("json", json.load),
            ("yaml", yaml.safe_load),
            ("yml", yaml.safe_load),
        ]:
            for file in config_path.rglob(f"*.{suffix}"):
                parts = file.relative_to(config_path).with_suffix("").parts
                with open(file) as f:
                    config_dict[".".join(parts)] = parser(f)

        return JSONPriorConfig(config_dict, directory=directory)

    def __str__(self):
        return json.dumps(self.obj)

    def __getitem__(self, item):
        return JSONPriorConfig(self.obj[".".join(item)], directory=self.directory)

    def __contains__(self, item):
        return ".".join(item) in self.obj

    def for_class_and_suffix_path(self, cls: Type, suffix_path: List[str]):
        """
        Get configuration for a prior.

        If it is just basic configuration then the suffix path is just the
        name of the prior in a list. Width configuration also adds an
        additional "width_modifier" item.

        If configuration is not found for the class then configurations for
        parents of the class are searched.

        Parameters
        ----------
        cls
            The class with which the prior is associated.
        suffix_path
            The path to the prior.

        Returns
        -------
        A configuration dictionary
        """
        for c in family(cls):
            try:
                return self(path_for_class(c) + suffix_path)
            except KeyError:
                pass
        raise KeyError(
            f"No config found for class {cls} and path {suffix_path} in {self.directory}"
        )

    def __call__(self, config_path: List[str]):
        """
        Get the config at the end of the config_path.

        The configuration dictionary is traversed until config is found
        at the end, else an exception is thrown.

        Parameters
        ----------
        config_path
            The import path of a package, module, class or class and constructor
            argument name.

        Returns
        -------
        A configuration dictionary or value

        Raises
        ------
        PriorException
            If no configuration is found.
        """
        key = ".".join(config_path)
        for path, value in self.path_value_tuples:
            if key.endswith(path):
                return value
        raise KeyError(
            f"No configuration was found for the path {config_path}"
            + ("" if self.directory is None else f" ({self.directory})")
        )



from functools import wraps
from typing import Callable
import logging

from autoconf.conf import instance

logger = logging.getLogger(__name__)


def should_output(name: str) -> bool:
    """
    Determine whether a file with a given name (excluding extension) should be output.

    This is configured in config/output.yaml. If the file is not present in the config, the default value is used.

    Parameters
    ----------
    name
        The name of the file to be output, excluding extension.

    Returns
    -------
    Whether the file should be output.
    """
    output_config = instance["output"]
    try:
        return output_config[name]
    except KeyError:
        return output_config["default"]


def conditional_output(func: Callable):
    """
    Decorator for functions that output files. If the file should not be output, the function is not called.

    Parameters
    ----------
    func
        A method where the first argument is the name of the file to be output.

    Returns
    -------
    The decorated function.
    """

    @wraps(func)
    def wrapper(self, name: str, *args, **kwargs):
        """
        Conditionally call the decorated function if the file should be output according
        to the config.

        Parameters
        ----------
        self
        name
            The name of the file to be output, excluding extension.
        args
        kwargs
        """
        if should_output(name):
            return func(self, name, *args, **kwargs)
        logger.info(f"Skipping output of {name}")

    return wrapper
